-module(dsdc_next_nonce_tests).

-include_lib("eunit/include/eunit.hrl").

-include("blocks.hrl").

-define(TEST_MODULE, dsdc_next_nonce).

-define(PUBKEY, <<"BAAggMEhrC3ODBqlYeQ6dk00F87AKMkV6kkyhgfJ/luOzGUC+4APxFkVgAYPai3TjSyLRObv0GeDACg1ZxwnfHY=">>).

pick_for_account_test_() ->
    {foreach,
     fun() ->
             meck:new(dsdc_chain),
             meck:new(dsdc_tx_pool)
     end,
     fun(_) ->
             meck:unload(dsdc_chain),
             meck:unload(dsdc_tx_pool)
     end,
     [{"Return account_not_found when both top block state tree and mempool are empty",
       fun() ->
               meck:expect(dsdc_chain, get_account, fun(?PUBKEY) -> none end),
               meck:expect(dsdc_tx_pool, get_max_nonce, fun(?PUBKEY) -> undefined end),
               ?assertEqual({error, account_not_found},
                            ?TEST_MODULE:pick_for_account(?PUBKEY))
       end},
      {"Return incremented state tree nonce for account existing in state tree and empty mempool",
       fun() ->
               Account = dsdc_accounts:set_nonce(dsdc_accounts:new(?PUBKEY, 0), 8),
               meck:expect(dsdc_chain, get_account, fun(?PUBKEY) -> {value, Account} end),
               meck:expect(dsdc_tx_pool, get_max_nonce, fun(?PUBKEY) -> undefined end),
               ?assertEqual({ok, 9}, ?TEST_MODULE:pick_for_account(?PUBKEY))
       end},
      {"Return [max(mempool account nonce, state tree account nonce) + 1] for account present in state and having txs in mempool",
       fun() ->
               Account = dsdc_accounts:set_nonce(dsdc_accounts:new(?PUBKEY, 0), 8),
               meck:expect(dsdc_chain, get_account, fun(?PUBKEY) -> {value, Account} end),
               meck:expect(dsdc_tx_pool, get_max_nonce, fun(?PUBKEY) -> {ok, 12} end),
               ?assertEqual({ok, 13}, ?TEST_MODULE:pick_for_account(?PUBKEY))
       end}]}.
