%%%-------------------------------------------------------------------
%%% @copyright (C) 2017, Dasudian Technologies
%%% @doc
%%%     Encode and decode data and function calls according to
%%%     Sophia-AEVM-ABI.
%%% @end
%%%
%%%-------------------------------------------------------------------
-module(dsdso_abi).
-define(HASH_SIZE, 32).

-export([create_calldata/3]).

-spec create_calldata(binary(), string(), string()) -> binary() | {error, argument_syntax_error}.

create_calldata(Contract, Function, Argument) ->
    %% TODO: check that function exists in cotract.
    FunctionHandle = encode_function(Contract, Function),
    case dsdso_constants:string(Argument) of
        {ok, ParsedArgument} ->
            encode_call(FunctionHandle, ParsedArgument);
        {error, _} ->
            {error, argument_syntax_error}
    end.

%% Use a tuple to pass multiple arguments. TODO: impossible to pass a single
%% tuple argument.
encode_call(FunctionHandle, {tuple, _, Elems}) ->
    encode_call1(FunctionHandle, Elems);
encode_call(FunctionHandle, Arg) ->
    encode_call1(FunctionHandle, [Arg]).

encode_call1(FunctionHandle, ArgumentAsts) ->
    Arguments = lists:map(fun ast_to_erlang/1, ArgumentAsts),
    Call = list_to_tuple([FunctionHandle | Arguments]),
    Data = dsdso_data:to_binary(Call),
    _ArgumentTypes = lists:map(fun get_type/1, Arguments),
    %% TODO: Verify that the type matches the function signature.
    Data.

ast_to_erlang({int, _, N}) -> N;
ast_to_erlang({string, _, Bin}) -> Bin;
ast_to_erlang({unit, _}) -> {};
ast_to_erlang({tuple, _, Elems}) ->
    list_to_tuple(lists:map(fun ast_to_erlang/1, Elems));
ast_to_erlang({list, _, Elems}) ->
    lists:map(fun ast_to_erlang/1, Elems).

encode_function(_Contract, Function) ->
     << <<X>> || X <- Function>>.

%% TODO: Handle all sophia data types.
get_type(I) when is_integer(I) -> int;
get_type(T) when is_tuple(T) ->
    ListOfTypes = [get_type(E) || E <- tuple_to_list(T)],
    {tuple, list_to_tuple(ListOfTypes)};
get_type(B) when is_binary(B) ->
    string;
get_type([]) -> nil;
get_type([E|_]) ->
    {list, get_type(E)}.





