# About this release

[This release][this-release] is focused on Smart Contracts & State Channels MVPs.
It:
* Adds wormhole functionality. This impacts consesus.
* Changes the cryptographic signing to use Curve25519. This means all old node/signing keys are not useable anymore (`key/key.pub` is replaced by `sign_key/sign_key.pub`). This impacts consensus.
* Enforces contiguous nonces for transactions from the same account. This impacts consensus.
* Enforces positive transaction fee for all transactions. This impacts consensus.
* Persists the pending transactions in the DB. This impacts the persisted DB.
* Improves the stability of the testnet.
* Adds functionality to smart contracts that allows for executing of non-trivial Solidity and Sophia smart contracts (impacts consensus):
  * Storing and retrieving the state of a smart contract (in a MPT).
  * Storing calls and return values in a separate MPT for one block.
  * Handling state in Sophia.
* Adds a first version of the Varna syntax and a compiler for that syntax.

[this-release]: https://github.com/dasudian/dsdin/releases/tag/v0.13.0

This release introduces backward incompatible changes in the chain format:
* After upgrading your node, you will not have your previous balance (do not attempt to keep your key pair from previous releases "v0.12.x" because not useable anymore);
* Please ensure that you do not reuse a persisted blockchain produced by the previous releases "v0.12.x".

Please join the testnet by following the instructions below, and let us know if you have any problems by [opening a ticket](https://github.com/dasudian/dsdin/issues).
Troubleshooting of common issues is documented [in the wiki](https://github.com/dasudian/dsdin/wiki/Troubleshooting).

The instructions below describe:
* [How to retrieve the released software for running a node](#retrieve-the-software-for-running-a-node);
* [How to install a node](#install-node);
* [How to join the testnet](#join-the-testnet).

## Retrieve the software for running a node

You can run a node by either:
* Installing the published [release binary][this-release] corresponding to your platform; or
* Running the published [Docker image `dasudian/dsdin`][docker]; or
* [Building a release binary from source][build].

[docker]: https://github.com/dasudian/dsdin/blob/v0.13.0/docs/docker.md
[build]: https://github.com/dasudian/dsdin/blob/v0.13.0/docs/build.md

The user configuration is documented in the [wiki](https://github.com/dasudian/dsdin/wiki/User-provided-configuration).
For specifying configuration using the Docker image, please refer to [its documentation][docker].

The node user API is documented:
* HTTP API endpoints are specified [online in swagger.yaml][swagger-yaml];
  * A JSON version of same specification is located in the node at path `lib/aehttp-0.1.0/priv/swagger.json`;
  * An interactive visualization of the same specification is available [online][swagger-ui].
* WebSocket API endpoints are [specified online][api-doc];
* The intended usage of the user API (HTTP and WebSocket) is [documented online][api-doc].

[swagger-yaml]: https://github.com/dasudian/dsdin/blob/v0.13.0/config/swagger.yaml
[swagger-ui]: https://dasudian.github.io/dsdin-api-docs/?config=https://raw.githubusercontent.com/dasudian/dsdin/v0.13.0/apps/aehttp/priv/swagger.json
[api-doc]: https://github.com/dasudian/protocol/blob/dsdin-v0.13.0/dsdin/api/README.md

## Install node

The instructions for installing a node using a release binary are in [the dedicated separate document](../../docs/installation.md).

For installation of a node using the Docker image, please refer to [its documentation online][docker].

## Join the testnet

This section describes how to run a node as part of the testnet - the public test network of nodes - by using the release binary.

For running a node as part of the testnet by using the Docker image, please consult [its documentation][docker] in addition to this section.

### Inspect the testnet

The core nodes of the public test network are accessible from the Internet.

Information, e.g. height, of the top block of the longest chain as seen by these core nodes of the testnet can be obtained by opening in the browser any of the following URLs:
* http://52.10.46.160:3013/v2/top
* http://18.195.109.60:3013/v2/top
* http://13.250.162.250:3013/v2/top
* http://31.13.249.70:3013/v2/top

### Setup your node

Setting up your node consists of:
* Configuring your node - see instructions in [the dedicated separate document](../../docs/configuration.md);
* Starting your node and verifying it works as expected - see instructions in [the dedicated separate document](../../docs/operation.md).
